from celery import shared_task


@shared_task
def update_indexed_document(index, created, pk):
    indexed_doc = index.init_using_pk(pk)
    indexed_doc.prepare()
    indexed_doc.save()


@shared_task
def delete_indexed_document(index, pk):
    indexed_doc = index.get(id=pk)
    indexed_doc.delete()


__all__ = ('update_indexed_document', 'delete_indexed_document')



# -*- coding: utf-8
import six

from importlib import import_module

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.apps import AppConfig

from . import connections, logger, flexconfig


class ElasticsearchFlexConfig(AppConfig):
    name = 'elasticsearch_flex'

    def ready(self):
        self.__init_connection()
        self.__import_modules()

    def __init_connection(self):
        host = flexconfig.get('host')
        if host is None:
            connections.create_connection(hosts=['localhost'])
            logger.info('No Elasticsearch host specified, assuming "localhost"')
            return

        if isinstance(host, dict):
            connections.configure(**host)
        elif isinstance(host, six.string_types):
            connections.create_connection(hosts=[host])
        else:
            raise ImproperlyConfigured('<host = {0}> for ElasticsearchFlex is incorrect'.format(host))

        logger.info('Elasticsearch connection configured using <%s>', host)

    def __import_modules(self):
        # Discover the modules
        import elasticsearch_flex.signals
        _loaded = []
        for app_name in settings.INSTALLED_APPS:
            module = '{}.search_indexes'.format(app_name)
            try:
                import_module(module)
                _loaded.append(app_name)
            except ImportError:
                pass
        if len(_loaded):
            logger.info('Loaded search indices for %s apps: %s', len(_loaded), _loaded)



from elasticsearch_dsl import field, analyzer

__base_fields = (
    'Object', 'String', 'Text', 'Date', 'Keyword', 'Nested',
    'InnerObjectWrapper',
    'Float', 'Double', 'Byte', 'Short', 'Integer', 'Long', 'Boolean',
    'Ip', 'Attachment',
    'GeoPoint', 'GeoShape',
    'Completion',
)


html_strip = analyzer(
    'html_strip',
    tokenizer="standard",
    filter=["standard", "lowercase", "stop", "snowball"],
    char_filter=["html_strip"]
)


class FlexField(object):
    def __init__(self, attr=None, *a, **kwa):
        self._model_attr = attr
        super(FlexField, self).__init__(*a, **kwa)


class HTMLField(FlexField, field.Text):
    def __init__(self, *a, **kwa):
        kwa['analyzer'] = html_strip
        super(HTMLField, self).__init__(*a, **kwa)


def __make_field(fieldname):
    class_name = fieldname + 'Field'
    base = getattr(field, fieldname)
    klass = type(class_name, (FlexField, base), {'__doc__': base.__doc__})
    return klass

__all__ = ['FlexField']
for __field in __base_fields:
    fclass = __make_field(__field)
    globals()[fclass.__name__] = fclass
    __all__.append(fclass.__name__)



'''Search Queryset

This search queryset wraps the low level elasticsearch queries, and acts as
a transparent interface between elasticsearch documents and corresponding
django model objects.
'''
from collections import namedtuple
from elasticsearch_dsl.search import Search

from . import connections


def from_queryset(hits, qs):
    ids = [hit._id for hit in hits]
    objects = qs.in_bulk(ids)
    keytype = type(objects.keys()[0])
    pks = list(map(keytype, ids))
    return pks, [objects[i] for i in pks if i in objects]


class ModelSearch(Search):
    def objects(self, qs=None):
        docs = list(self)
        if len(docs):
            if qs is None:
                # Use one document for introspection.
                # This resolves the default queryset defined with index.
                qs = docs[0].get_queryset()
            _, objects = from_queryset(docs, qs)
            return objects
        return []


class TemplateSearch(ModelSearch):
    def execute(self, ignore_cache=False):
        if ignore_cache or not hasattr(self, '_response'):
            es = connections.get_connection(self._using)
            payload = {
                'id': self._extra['t_name'],
                'params': self._params,
            }

            self._response = self._response_class(
                self,
                es.search_template(
                    index=self._index,
                    doc_type=self._doc_type,
                    body=payload,
                )
            )
        return self._response


class DocAccessors(object):
    def __init__(self, index):
        self.index = index

    @property
    def dsl(self):
        return ModelSearch().doc_type(self.index)

    @property
    def templates(self):
        index_name = self.index._meta.index
        available_templates = getattr(self.index._meta, 'query_templates', [])

        FlexTemplates = namedtuple('FlexTemplates', available_templates)
        dsl = []
        for _id in available_templates:
            tid = index_name + '.' + _id
            s = TemplateSearch(index=index_name, extra={'t_name': tid}, doc_type=self.index)
            dsl.append(s)

        return FlexTemplates(*dsl)
